import pandas as pd
from idcempy import zmiopc
import os
import unittest
import numpy as np

"""
Replication of results from 
Bagozzi, Benjamin E. and Bumba Mukherjee.  2012. 
"A Mixture Model for Middle-category Inflation in Ordered Survey Responses."  
Political Analysis. 20(3): 369-386
"""
DAT = pd.read_stata(
    os.getcwd() + "/data/EUKnowledge.dta", convert_categoricals=False)

Y = ["EU_support_ET"]

X = ["polit_trust", "Xenophobia", "discuss_politics", "Professional",
     "Executive", "Manual", "Farmer", "Unemployed", "rural", "female", "age",
     "student", "income", "Educ_high", "Educ_high_mid", "Educ_low_mid"]

Z = ["discuss_politics", "rural", "female", "age",
     "student", "EUbid_Know", "EU_Know_obj", "TV", "Educ_high",
     "Educ_high_mid", "Educ_low_mid"]

varlist = np.unique(Y + Z + X)
dataset = DAT[varlist]
datasetnew = dataset.dropna(how="any")
datasetnew = datasetnew.reset_index(drop=True)
x_ = datasetnew[X]
y_ = datasetnew[Y]
yx_ = y_.iloc[:, 0]
yncat = len(np.unique(yx_))
z_ = datasetnew[Z]
z_.insert(0, "ones", np.repeat(1, len(z_)))

pstartmiopc = np.repeat(
    0.01, ((yncat - 1) + len(x_.columns) + len(z_.columns) + 1)
)

pstartmiop = np.repeat(0.01, (
        (yncat - 1) + len(x_.columns) + len(z_.columns)))


# miopc_EU = zmiopc.iopcmod('miopc', DAT, X, Y, Z)
# miop_EU = zmiopc.iopmod('miop', DAT, X, Y, Z)
# op_EU = zmiopc.opmod(DAT, X, Y)
class TestMiopLlike(unittest.TestCase):
    def test_zioploglike(self):
        self.assertAlmostEqual(zmiopc.miop(pstartmiop, x_, yx_, z_,
                                           datasetnew, 1, 0, 0), 11030.403,
                               places=2)


class TestMiopCLlike(unittest.TestCase):
    def test_ziopcloglike(self):
        self.assertAlmostEqual(zmiopc.miopc(pstartmiopc, x_, yx_, z_,
                                            datasetnew, 1, 0, 0), 11009.157,
                               places=2)


class TestOp(unittest.TestCase):
    def test_opmodel(self):
        self.assertAlmostEqual(zmiopc.opmod(DAT, X, Y).coefs.iloc[2, 0],
                               0.76, places=1)


class TestMiop(unittest.TestCase):
    def test_miopmodel(self):
        self.assertAlmostEqual(zmiopc.iopmod('miop',
                                             DAT, X, Y, Z).coefs.iloc[2, 0],
                               0.43, places=1)
        self.assertAlmostEqual(zmiopc.iopmod('miop',
                                             DAT, X, Y, Z).coefs.iloc[14, 0],
                               0.90, places=1)


class TestMiopc(unittest.TestCase):
    def test_miopcmodel(self):
        self.assertAlmostEqual(zmiopc.iopcmod('miopc',
                                              DAT, X, Y, Z).coefs.iloc[14, 0],
                               0.84, places=1)



import pandas as pd
from idcempy import gimnl
import os
import unittest
import numpy as np

"""
Replication of results from 
Bagozzi, Benjamin E. and Kathleen Marchetti. 2017. 
"Distinguishing Occasional Abstention 
rom Routine Indifference in Models of Vote Choice." 
Political Science Research  and Methods.  5(2): 277-249.
"""

DAT = pd.read_stata(
    os.getcwd() + "/data/replicationdata.dta", convert_categoricals=False)

x = ['educ', 'female', 'black', 'hispanic', 'party7', 'w3gbvalu2',
     'presbatt',
     'south', 'gayban2', 'agegroup2', 'bornagn_w', 'brnag_wXgmb2', 'catholic',
     'cathXgmb2', 'other_rel', 'secular', 'secXgmb2', 'ideo', 'w3mobidx']
z = ['educ', 'agegroup2', 'w3mobidx', 'secular']
y = ['vote_turn']

order_Kerry = [0, 1, 2]
order_Bush = [0, 2, 1]

binflatecat = "baseline"
sinflatecat = "second"
tinflatecat = "third"

varlist = np.unique(y + z + x)
dataset = DAT[varlist]
datasetnew = dataset.dropna(how="any")
datasetnew = datasetnew.reset_index(drop=True)
x_ = datasetnew[x]
y_ = datasetnew[y]
yx_ = y_.iloc[:, 0]
yncat = len(np.unique(yx_))
z_ = datasetnew[z]
z_.insert(0, "int", np.repeat(1, len(z_)))
x_.insert(0, "int", np.repeat(1, len(x_)))

pstart_mnl = np.repeat(0.01, len(x_.columns) + len(x_.columns))
pstart_gimnl = np.repeat(0.01, len(x_.columns) + len(x_.columns)
                         + len(z_.columns))


class TestBimnlLlike(unittest.TestCase):
    def test_bimnloglike(self):
        self.assertEqual(gimnl.bimnl3(pstart_gimnl, x_, x_, yx_, z_,
                                      [0, 1, 2]),
                         gimnl.bimnl3(pstart_gimnl, x_, x_, yx_, z_,
                                      [0, 2, 1]))
        self.assertAlmostEqual(gimnl.bimnl3(pstart_gimnl, x_, x_, yx_, z_,
                                            [0, 1, 2]), 2109.38643, places=2)


class TestSimnlTimnlLlike(unittest.TestCase):
    def test_stimnloglike(self):
        self.assertEqual(gimnl.simnl3(pstart_gimnl, x_, x_, yx_, z_,
                                      [2, 0, 1]),
                         gimnl.timnl3(pstart_gimnl, x_, x_, yx_, z_,
                                      [2, 1, 0]))
        self.assertAlmostEqual(gimnl.simnl3(pstart_gimnl, x_, x_, yx_, z_,
                                            [2, 0, 1]), 2242.670607, places=2)
        self.assertEqual(gimnl.simnl3(pstart_gimnl, x_, x_, yx_, z_,
                                      [1, 0, 2]),
                         gimnl.timnl3(pstart_gimnl, x_, x_, yx_, z_,
                                      [1, 2, 0]))
        self.assertAlmostEqual(gimnl.timnl3(pstart_gimnl, x_, x_, yx_, z_,
                                            [1, 2, 0]), 2242.880607, places=2)


class TestMnlLlike(unittest.TestCase):
    def test_mnloglike(self):
        self.assertEqual(gimnl.mnl3(pstart_mnl, x_, x_, yx_, [0, 1, 2]),
                         gimnl.mnl3(pstart_mnl, x_, x_, yx_, [0, 2, 1]))
        self.assertAlmostEqual(gimnl.mnl3(pstart_mnl, x_, x_, yx_, [0, 1, 2]),
                               1401.98008,
                               places=2)


class TestMnl(unittest.TestCase):
    def test_mnlmodel(self):
        self.assertEqual(len(
            gimnl.mnlmod(DAT, x, y, order_Bush).coefs),
            40)
        self.assertAlmostEqual(gimnl.mnlmod(
            DAT, x, y, order_Kerry).coefs.iloc[0, 0], -2.189,
                               places=1)
        self.assertAlmostEqual(gimnl.mnlmod(
            DAT, x, y, order_Kerry).coefs.iloc[5, 0], 0.344,
                               places=1)
        self.assertAlmostEqual(gimnl.mnlmod(
            DAT, x, y, order_Bush).coefs.iloc[0, 0], -3.018,
                               places=1)
        self.assertAlmostEqual(gimnl.mnlmod(
            DAT, x, y, order_Bush).coefs.iloc[5, 0], -0.386,
                               places=1)


class TestGimnl(unittest.TestCase):
    def test_gimmnlmodel(self):
        self.assertEqual(len(
            gimnl.gimnlmod(DAT, x, y, z, order_Kerry, "baseline").coefs),
            45)
        self.assertAlmostEqual(gimnl.gimnlmod(
            DAT, x, y, z, order_Bush, "baseline").coefs.iloc[0, 0], -4.938,
                               places=1)
        self.assertAlmostEqual(gimnl.gimnlmod(
            DAT, x, y, z, order_Bush, "baseline").coefs.iloc[5, 0], -1.632,
                               places=1)
        self.assertAlmostEqual(gimnl.gimnlmod(
            DAT, x, y, z, order_Bush, "baseline").coefs.iloc[10, 0], -0.390,
                               places=1)
        self.assertAlmostEqual(gimnl.gimnlmod(
            DAT, x, y, z, order_Kerry, "baseline").coefs.iloc[0, 0], -4.938,
                               places=1)
        self.assertAlmostEqual(gimnl.gimnlmod(
            DAT, x, y, z, order_Kerry, "baseline").coefs.iloc[5, 0], -0.724,
                               places=1)
        self.assertAlmostEqual(gimnl.gimnlmod(
            DAT, x, y, z, order_Kerry, "baseline").coefs.iloc[10, 0], 0.344,
                               places=1)


class TestVuongGimnl(unittest.TestCase):
    def test_vuongmnlgimnl(self):
        self.assertAlmostEqual(gimnl.vuong_gimnl(gimnl.mnlmod(
            DAT, x, y, order_Kerry), gimnl.gimnlmod(
            DAT, x, y, z, order_Kerry, "baseline")), -1.9174562877, places=4)



import pandas as pd
from idcempy import zmiopc
import os
import unittest
import numpy as np

"""
Replication of results from 
Bagozzi, Benjamin E., Daniel W. Hill Jr., Will H. Moore, and Bumba Mukherjee .
 2015. "Modeling Two Types of Peace: The Zero-inflated Ordered Probit (ZiOP) 
 Model in Conflict Research."  Journal of Conflict Resolution. 59(4): 728-752.
"""

DAT = pd.read_stata(
    os.getcwd() + "/data/bp_exact_for_analysis.dta",
    convert_categoricals=False)

# Ziop and ziopc examples
# Specifying Xs, Zs, and Y
X = ['logGDPpc', 'parliament', 'disaster', 'major_oil', 'major_primary']
Z = ['logGDPpc', 'parliament']
Y = ['rep_civwar_DV']
data = DAT

pstartziop = np.array([-1.31, .32, 2.5, -.21, .2, -0.2, -0.4, 0.2, .9, -.4])

pstartziopsmall = np.array([-1.31, .32, 2.5, -.21, .2, -0.2, -0.4, 0.2])

pstartziopc = np.array([-1.31, .32, 2.5, -.21,
                        .2, -0.2, -0.4, 0.2, .9, -.4, .1])

pstartop = np.array([-1, 0.3, -0.2, -0.5, 0.2, .9, -.4])

varlist = np.unique(Y + Z + X)
dataset = data[varlist]
datasetnew = dataset.dropna(how="any")
datasetnew = datasetnew.reset_index(drop=True)
x_ = datasetnew[X]
y_ = datasetnew[Y]
yx_ = y_.iloc[:, 0]
yncat = len(np.unique(yx_))
z_ = datasetnew[Z]
z_.insert(0, "ones", np.repeat(1, len(z_)))


class TestZiopLlike(unittest.TestCase):
    def test_zioploglike(self):
        self.assertAlmostEqual(zmiopc.ziop(pstartziop, x_, yx_, z_,
                                           datasetnew, 1, 0, 0), 1486.169,
                               places=2)


class TestZiopCLlike(unittest.TestCase):
    def test_ziopcloglike(self):
        self.assertAlmostEqual(zmiopc.ziopc(pstartziopc, x_, yx_, z_,
                                           datasetnew, 1, 0, 0), 1487.748,
                               places=2)


class TestOp(unittest.TestCase):
    def test_opmodel(self):
        self.assertAlmostEqual(zmiopc.opmod(data, X, Y,
                                            pstart=pstartop, method='bfgs',
                                            weights=1,
                                            offsetx=0).coefs.iloc[2, 0],
                               -0.21, places=1)


class TestZiop(unittest.TestCase):
    def test_ziopmodel(self):
        self.assertAlmostEqual(zmiopc.iopmod('ziop',
                                             data, X, Y, Z, pstart=pstartziop,
                                             method='bfgs', weights=1,
                                             offsetx=0,
                                             offsetz=0).coefs.iloc[4, 0],
                               -0.29, places=1)
        self.assertAlmostEqual(zmiopc.iopmod('ziop',
                                             data, X, Y, Z, pstart=pstartziop,
                                             method='bfgs', weights=1,
                                             offsetx=0,
                                             offsetz=0).coefs.iloc[5, 0],
                               0.04, places=1)


class TestZiopc(unittest.TestCase):
    def test_ziopcmodel(self):
        self.assertAlmostEqual(zmiopc.iopcmod('ziopc',
                                              data, X, Y, Z,
                                              pstart=pstartziopc,
                                              method='bfgs', weights=1,
                                              offsetx=0,
                                              offsetz=0).coefs.iloc[-1, 0],
                               -0.889, places=1)
        self.assertAlmostEqual(zmiopc.iopcmod('ziopc',
                                              data, X, Y, Z,
                                              pstart=pstartziopc,
                                              method='bfgs', weights=1,
                                              offsetx=0,
                                              offsetz=0).coefs.iloc[4, 0],
                               -0.37, places=1)
        self.assertAlmostEqual(zmiopc.iopcmod('ziopc',
                                              data, X, Y, Z,
                                              pstart=pstartziopc,
                                              method='bfgs', weights=1,
                                              offsetx=0,
                                              offsetz=0).coefs.iloc[5, 0],
                               0.33, places=1)


class VuongOpZiopc(unittest.TestCase):
    def test_vuongopziopc(self):
        self.assertAlmostEqual(zmiopc.vuong_opiop(zmiopc.opmod(data, X, Y,
                                                               pstart=pstartop,
                                                               method='bfgs',
                                                               weights=1,
                                                               offsetx=0),
                                                  zmiopc.iopmod('ziop',
                                                                data, X, Y, Z,
                                                                pstart=
                                                                pstartziop,
                                                                method='bfgs',
                                                                weights=1,
                                                                offsetx=0,
                                                                offsetz=0)),
                               -4.909, places=1)
        self.assertAlmostEqual(zmiopc.vuong_opiopc(zmiopc.opmod(data, X, Y,
                                                                pstart=
                                                                pstartop,
                                                                method='bfgs',
                                                                weights=1,
                                                                offsetx=0),
                                                   zmiopc.iopcmod('ziopc',
                                                                  data, X, Y,
                                                                  Z,
                                                                  pstart=
                                                                  pstartziopc,
                                                                  method='bfgs'
                                                                  , weights=1,
                                                                  offsetx=0,
                                                                  offsetz=0)),
                               -5.424, places=1)



"""Classes and Functions for the ziopcpy module."""
import numpy as np

# ZiOPC model converges extremely
# faster with import * rather than import as np.
import pandas as pd
from numpy import *
from scipy.optimize import minimize
from scipy.stats import mvn, norm


class OpModel:
    """Store model results from :py:func:`opmod`."""

    def __init__(self, llik, coef, aic, vcov, data, xs, ts, x_, yx_, yncat,
                 xstr, ystr):
        """Store model results, goodness-of-fit tests, and other information.

        :param llik: Log-Likelihood.
        :param coef: Model coefficients.
        :param aic: Model Akaike information criterion.
        :param vcov: Variance-Covariance matrix.
            (optimized as inverted Hessian matrix)
        :param data: Full dataset.
        :param ts: Cutpoints for ordered probit.
        :param xs: Ordered probit estimates (Betas).
        :param yncat: Number of categories in the outcome variable.
        :param x_: X (Covariates) Data.
        :param yx_: Y (Dependent Variable) data.
        :param xstr: List of strings for variable(s) in the outcome stage (x).
        :param ystr: List of strings for outcome variable name (y).
        """
        self.llik = llik
        self.coefs = coef
        self.AIC = aic
        self.vcov = vcov
        self.data = data
        self.cutpoints = ts
        self.ordered = xs
        self.ycat = yncat
        self.X = x_
        self.Y = yx_
        self.xstr = xstr
        self.ystr = ystr


class IopModel:
    """Store model results from :py:func:`iopmod`."""

    def __init__(
            self,
            modeltype,
            llik,
            coef,
            aic,
            vcov,
            data,
            xs,
            zs,
            ts,
            x_,
            yx_,
            z_,
            yncat,
            xstr,
            ystr,
            zstr,
    ):
        """Store model results, goodness-of-fit tests, and other information.

        :param modeltype: Type of Iop Model (ziop or miop).
        :param llik: Log-Likelihood.
        :param coef: Model coefficients
        :param aic: Model Akaike information criterion.
        :param vcov: Variance-Covariance matrix.
            (optimized as inverted Hessian matrix)
        :param data: Full dataset.
        :param ts: Cutpoints for ordered probit.
        :param zs: Inflation stage estimates (Gammas).
        :param xs: Ordered probit estimates (Betas).
        :param yncat: Number of categories in the Dependent Variable (DV).
        :param x_: X Data.
        :param yx_: Y (DV) data.
        :param z_: Z Data.
        :param xstr: List of strings for x names.
        :param ystr: List of strings for y names.
        :param zstr: List of strings for z names.

        """
        self.modeltype = modeltype
        self.llik = llik
        self.coefs = coef
        self.AIC = aic
        self.vcov = vcov
        self.data = data
        self.cutpoints = ts
        self.inflate = zs
        self.ordered = xs
        self.ycat = yncat
        self.X = x_
        self.Y = yx_
        self.Z = z_
        self.xstr = xstr
        self.ystr = ystr
        self.zstr = zstr


class IopCModel:
    """Store model results from :py:func:`iopcmod`."""

    def __init__(
            self,
            modeltype,
            llik,
            coef,
            aic,
            vcov,
            data,
            xs,
            zs,
            ts,
            x_,
            yx_,
            z_,
            rho,
            yncat,
            xstr,
            ystr,
            zstr,
    ):
        """Store model results, goodness-of-fit tests, and other information.

        :param modeltype: Type of IopC Model (ziopc or miopc).
        :param llik: Log-Likelihood.
        :param coef: Model coefficients.
        :param AIC: Model Akaike information criterion.
        :param vcov: Variance-Covariance matrix.
            (optimized as inverted Hessian matrix)
        :param data: Full dataset.
        :param ts: Cutpoints for ordered probit.
        :param zs: Inflation stage estimates (Gammas).
        :param xs: Ordered probit estimates (Betas).
        :param rho: Rho.
        :param yncat: Number of categories in Dependent Variable (DV).
        :param x_: X Data.
        :param yx_: Y (DV) data.
        :param z_: Z Data.
        :param xstr: List of strings for x names.
        :param ystr: List of strings for y names.
        :param zstr: List of strings for z names.
        """
        self.modeltype = modeltype
        self.llik = llik
        self.coefs = coef
        self.AIC = aic
        self.vcov = vcov
        self.data = data
        self.cutpoints = ts
        self.inflate = zs
        self.ordered = xs
        self.ycat = yncat
        self.X = x_
        self.Y = yx_
        self.Z = z_
        self.r = rho
        self.xstr = xstr
        self.ystr = ystr
        self.zstr = zstr


class FittedVals:
    """Store fitted values for iOP models."""

    def __init__(self, responsefull, responseordered, responseinflation,
                 linear):
        """Store different type of equation in each attribute.

        :param responsefull: responsefull
        :param responseordered: responseordered
        :param responseinflation: responseinflation
        :param linear: linear
        """
        self.responsefull = responsefull
        self.responseordered = responseordered
        self.responseinflation = responseinflation
        self.linear = linear


def op(pstart, x, y, data, weights, offsetx):
    """Calculate likelihood function for Ordered Probit Model.

    :param pstart: A list of starting values for the estimation.
        Length of the number of parameters to be estimated.
    :type pstart: list
    :param x: Covariates for the ordered stage.
    :type x: pandas.DataFrame
    :param y: The dependent variable (DV).
    :type y: pandas.DataFrame
    :param data: Dataset.
    :type data: pandas.DataFrame
    :param weights: Weights.
    :param offsetx: Offset for covariates in the ordered stage.
    """
    n = len(data)
    ycat = y.astype("category")
    ycatu = np.unique(ycat)
    yncat = len(ycatu)
    y0 = np.sort(ycatu)
    v = np.zeros((len(data), yncat))
    for j in range(yncat):
        v[:, j] = y == y0[j]
    tau = np.repeat(1.0, yncat)
    for j in range(yncat - 1):
        if j == 0:
            tau[j] = pstart[j]
        else:
            tau[j] = tau[j - 1] + np.exp(pstart[j])
    beta = pstart[(yncat - 1): len(pstart)]
    xb = x.dot(beta) + offsetx
    cprobs = np.zeros((n, yncat))
    probs = np.zeros((n, yncat))
    for i in range(yncat - 1):
        cprobs[:, i] = norm.cdf(tau[i] - xb)
    probs[:, 0] = cprobs[:, 0]
    for i in range(1, yncat - 1):
        probs[:, i] = cprobs[:, i] - cprobs[:, (i - 1)]
    probs[:, yncat - 1] = 1 - cprobs[:, (yncat - 2)]
    lik = np.zeros((n, yncat))
    for k in range(n):
        for j in range(yncat):
            lik[k, j] = v[k, j] * probs[k, j]
    likk = np.log(lik[lik != 0])
    llik = -1 * sum(likk * weights)
    return llik


def ziop(pstart, x, y, z, data, weights, offsetx, offsetz):
    """Calculate likelihood function for Zero-inflated Model.

    :param pstart: A list of starting values for the estimation.
        Length of the number of parameters to be estimated.
    :type pstart: list
    :param x: Covariates for the ordered probit stage. 
    :type x: pandas.DataFrame
    :param y: The ordinal dependent Variable (DV). 
    :type y: pandas.DataFrame
    :param z: Covariates for the inflation stage. 
    :type z: pandas.DataFrame
    :param data: Dataset with missing values listwise deleted.
    :type data: pandas.DataFrame
    :param weights: weights.
    :type weights: float
    :param offsetx: Offset for the ordered stage covariates (X). 
    :type offsetx: float
    :param offsetz: Offset for the inflation stage covariates (Z).
    :type offsetz: float
    """
    n = len(data)
    ycat = y.astype("category")
    ycatu = np.unique(ycat)
    yncat = len(ycatu)
    y0 = np.sort(ycatu)
    v = np.zeros((len(data), yncat))
    for j in range(yncat):
        v[:, j] = y == y0[j]
    tau = np.repeat(1.0, yncat)
    for j in range(yncat - 1):
        if j == 0:
            tau[j] = pstart[j]
        else:
            tau[j] = tau[j - 1] + np.exp(pstart[j])
    beta = pstart[(yncat + len(z.columns) - 1): len(pstart)]
    gamma = pstart[(yncat - 1): (yncat + len(z.columns) - 1)]
    zg = z.dot(gamma) + offsetz
    xb = x.dot(beta) + offsetx
    cprobs = np.zeros((n, yncat))
    probs = np.zeros((n, yncat))
    for i in range(yncat - 1):
        cprobs[:, i] = norm.cdf(tau[i] - xb)
    probs[:, 0] = (cprobs[:, 0]) * norm.cdf(zg) + (1 - norm.cdf(zg))
    for i in range(1, yncat - 1):
        probs[:, i] = (cprobs[:, i] - cprobs[:, (i - 1)]) * norm.cdf(zg)
    probs[:, yncat - 1] = (1 - cprobs[:, (yncat - 2)]) * norm.cdf(zg)
    lik = np.zeros((n, yncat))
    for k in range(n):
        for j in range(yncat):
            lik[k, j] = v[k, j] * probs[k, j]
    likk = np.log(lik[lik != 0])
    llik = -1 * sum(likk * weights)
    return llik


def ziopc(pstart, x, y, z, data, weights, offsetx, offsetz):
    """Calculate likelihood function for Zero-inflated Correlated-Errors Model.

    :param pstart: A list of starting values for the estimation.
        Length of the number of parameters to be estimated.
    :type pstart: list
    :param x: Covariates for the ordered probit stage. 
    :type x: pandas.DataFrame
    :param y: The dependent variable (DV).
    :type y: pandas.DataFrame
    :param z: Covariates for the inflation stage. 
    :type z: pandas.DataFrame
    :param data: Dataset with missing values listwise deleted.
    :type data: pandas.DataFrame
    :param weights: Weights.
    :type weights: float
    :param offsetx: Offset for the ordered probit stage covariates (X).
    :type offsetx: float
    :param offsetz: Offset for the inflation stage (Z).
    :type offsetz: float
    """
    n = len(data)
    ycat = y.astype("category")
    ycatu = np.unique(ycat)
    yncat = len(ycatu)
    y0 = np.sort(ycatu)
    v = np.zeros((len(data), yncat))
    for j in range(yncat):
        v[:, j] = y == y0[j]
    tau = np.repeat(1.0, yncat)
    for j in range(yncat - 1):
        if j == 0:
            tau[j] = pstart[j]
        else:
            tau[j] = tau[j - 1] + np.exp(pstart[j])
    beta = pstart[(yncat + len(z.columns) - 1): len(pstart) - 1]
    gamma = pstart[(yncat - 1): (yncat + len(z.columns) - 1)]
    x_beta = x.dot(beta)
    # correlation
    rho = pstart[len(pstart) - 1]
    cprobs = np.zeros((len(x_beta), yncat))
    probs = np.zeros((len(x_beta), yncat))
    cut = np.zeros((len(x_beta), yncat))
    cutb = np.zeros((len(x_beta), yncat))
    zg = z.dot(gamma) + offsetz
    xb = x.dot(beta) + offsetx
    means = np.array([0, 0])
    lower = np.array([-inf, -inf])
    sigma = np.array([[1, rho], [rho, 1]])
    nsigma = np.array([[1, -rho], [-rho, 1]])
    for i in range(yncat - 1):
        cprobs[:, i] = norm.cdf(tau[i] - xb)
        cut[:, i] = tau[i] - xb
        cutb[:, i] = xb - tau[i]
    upperb = np.zeros((len(x_beta), 2))
    upper = np.zeros((len(x_beta), 2))
    for j in range(n):
        upperb[j, :] = [zg[j], cutb[j, yncat - 2]]
        upper[j, :] = [zg[j], cut[j, 0]]
        probs[j, yncat - 1] = mvn.mvnun(lower, upperb[j], means, sigma)[0]
        probs[j, 0] = (1 - norm.cdf(zg[j])) + mvn.mvnun(lower, upper[j],
                                                        means, nsigma)[0]
    for j in range(n):
        for i in range(1, yncat - 1):
            probs[j, i] = (
                    mvn.mvnun(lower, [zg[j],
                                      cut[j, i]], means, nsigma)[0]
                    - mvn.mvnun(lower, [zg[j],
                                        cut[j, i - 1]], means, nsigma)[0])
    lik = np.zeros((n, yncat))
    for k in range(n):
        for j in range(yncat):
            lik[k, j] = v[k, j] * probs[k, j]
    likk = np.log(lik[lik != 0])
    llik = -1 * sum(likk * weights)
    return llik


def miop(pstart, x, y, z, data, weights, offsetx, offsetz):
    """
    Likelihood function for Middle-inflated Ordered Probit Model
    "without" correlated errors.

    The number of categories in the dependent
    variable must be odd.

    :param pstart: A list of starting values for the estimation.
        Length of the number of parameters to be estimated.
    :type pstart: list
    :param x: Covariates in the ordered probit stage. 
    :type x: pandas.DataFrame
    :param y: The dependent variable (DV).
    :type y: pandas.DataFrame
    :param z: Covariates in the inflation stage. 
    :type z: pandas.DataFrame
    :param data: Dataset with missing values listwise deleted.
    :type data: pandas.DataFrame
    :param weights: Weights.
    :type weights: float
    :param offsetx: Offset for covariates in the ordered probit stage (X).
    :type offsetx: float
    :param offsetz: Offset for covariates in the inflation stage.
    :type offsetz: float
    """
    n = len(data)
    ycat = y.astype("category")
    ycatu = np.unique(ycat)
    yncat = len(ycatu)
    y0 = np.sort(ycatu)
    v = np.zeros((len(data), yncat))
    for j in range(yncat):
        v[:, j] = y == y0[j]
    tau = np.repeat(1.0, yncat)
    for j in range(yncat - 1):
        if j == 0:
            tau[j] = pstart[j]
        else:
            tau[j] = tau[j - 1] + np.exp(pstart[j])
    beta = pstart[(yncat + len(z.columns) - 1): len(pstart)]
    gamma = pstart[(yncat - 1): (yncat + len(z.columns) - 1)]
    zg = z.dot(gamma) + offsetz
    xb = x.dot(beta) + offsetx
    cprobs = np.zeros((n, yncat))
    probs = np.zeros((n, yncat))
    for i in range(yncat - 1):
        cprobs[:, i] = norm.cdf(tau[i] - xb)
    probs[:, 0] = (cprobs[:, 0]) * norm.cdf(zg)
    for i in range(1, yncat - 1):
        if i == np.median(range(yncat)):
            probs[:, i] = (1 - norm.cdf(zg)) + (
                    norm.cdf(zg) * (cprobs[:, i] - cprobs[:, (i - 1)])
            )
        else:
            probs[:, i] = norm.cdf(zg) * (cprobs[:, i] - cprobs[:, (i - 1)])
    probs[:, yncat - 1] = (1 - cprobs[:, (yncat - 2)]) * norm.cdf(zg)
    lik = np.zeros((n, yncat))
    for k in range(n):
        for j in range(yncat):
            lik[k, j] = v[k, j] * probs[k, j]
    likk = np.log(lik[lik != 0])
    llik = -1 * sum(likk * weights)
    return llik


def miopc(pstart, x, y, z, data, weights, offsetx, offsetz):
    """
    Likelihood function for Middle-inflated Correlated-Errors Model.
    
    The number of categories in the dependent variable must be odd.

    :param pstart: A list of starting values for the estimation.
        Length of the number of parameters to be estimated.
    :type pstart: list
    :param x: Covariates in the ordered probit stage.
    :type x: pandas.DataFrame
    :param y: The dependent variable (DV). 
    :type y: pandas.DataFrame
    :param z: Covariates in the inflation stage.
    :type z: pandas.DataFrame
    :param data: Dataset with missing values listwise deleted.
    :type data: pandas.DataFrame
    :param weights: Weights.
    :type weights: float
    :param offsetx: Offset for the ordered probit stage covariates (X).
    :type offsetx: float
    :param offsetz: Offset for the covariates in the inflation stage (Z).
    :type offsetz: float
    """
    n = len(data)
    ycat = y.astype("category")
    ycatu = np.unique(ycat)
    yncat = len(ycatu)
    y0 = np.sort(ycatu)
    v = np.zeros((len(data), yncat))
    for j in range(yncat):
        v[:, j] = y == y0[j]
    tau = np.repeat(1.0, yncat)
    for i in range(yncat - 1):
        if i == 0:
            tau[i] = pstart[i]
        else:
            tau[i] = tau[i - 1] + np.exp(pstart[i])
    beta = pstart[(yncat + len(z.columns) - 1): len(pstart) - 1]
    gamma = pstart[(yncat - 1): (yncat + len(z.columns) - 1)]
    x_beta = x.dot(beta)
    rho = pstart[len(pstart) - 1]
    cprobs = np.zeros((len(x_beta), yncat))
    probs = np.zeros((len(x_beta), yncat))
    cutpoint = np.zeros((len(x_beta), yncat))
    cutpointb = np.zeros((len(x_beta), yncat))
    zg = z.dot(gamma) + offsetz
    xb = x.dot(beta) + offsetx
    means = np.array([0, 0])
    lower = np.array([-inf, -inf])
    sigma = np.array([[1, rho], [rho, 1]])
    nsigma = np.array([[1, -rho], [-rho, 1]])
    for i in range(yncat - 1):
        cprobs[:, i] = norm.cdf(tau[i] - xb)
        cutpoint[:, i] = tau[i] - xb
        cutpointb[:, i] = xb - tau[i]
    upperb = np.zeros((len(x_beta), 2))
    upper = np.zeros((len(x_beta), 2))
    for j in range(n):
        upperb[j, :] = [zg[j], cutpointb[j, yncat - 2]]
        upper[j, :] = [zg[j], cutpoint[j, 0]]
        probs[j, yncat - 1] = mvn.mvnun(lower, upperb[j], means, sigma)[0]
        probs[j, 0] = mvn.mvnun(lower, upper[j], means, nsigma)[0]
    for i in range(n):
        for j in range(1, yncat - 1):
            if j == np.median(range(yncat)):
                probs[i, j] = (
                        (1 - norm.cdf(zg[i]))
                        + mvn.mvnun(lower, [zg[i], cutpoint[i, j]], means,
                                    nsigma)[0]
                        - mvn.mvnun(lower, [zg[i], cutpoint[i, j - 1]], means,
                                    nsigma)[0]
                )
            else:
                probs[i, j] = (
                        mvn.mvnun(lower, [zg[i], cutpoint[i, j]], means,
                                  nsigma)[0]
                        - mvn.mvnun(lower, [zg[i], cutpoint[i, j - 1]], means,
                                    nsigma)[0]
                )
    lik = np.zeros((n, yncat))
    for k in range(n):
        for j in range(yncat):
            lik[k, j] = v[k, j] * probs[k, j]
    likk = np.log(lik[lik != 0])
    llik = -1 * sum(likk * weights)
    return llik


def opresults(model, data, x, y):
    """Produces estimation results, part of :py:func:`opmod`.

    :param model: Model estimation results obtained from minimization.
    :param data: Dataset.
    :type data: pandas.DataFrame
    :param x: List of names for independent variables matching column names
        in data.
    :type x: list of str
    :param y: : List of name for dependent Variablematching column names
        in data.
    """
    varlist = np.unique(y + x)
    dataset = data[varlist]
    datasetnew = dataset.dropna(how="any")
    datasetnew = datasetnew.reset_index(drop=True)
    x_ = datasetnew[x]
    y_ = datasetnew[y]
    yx_ = y_.iloc[:, 0]
    yncat = len(np.unique(yx_))
    names = list()
    for s in range(1, yncat):
        names.append("cut" + str(s))
    for s in range(x_.shape[1]):
        names.append(x_.columns[s])
    ts = model.x[0: yncat - 1]
    xs = model.x[(yncat - 1): (yncat + x_.shape[1] - 1)]
    ses = np.sqrt(np.diag(model.hess_inv))
    tscore = model.x / ses
    pval = (1 - (norm.cdf(abs(tscore)))) * 2
    lci = model.x - 1.96 * ses
    uci = model.x + 1.96 * ses
    coef = pd.DataFrame(
        {
            "Coef": model.x,
            "SE": ses,
            "tscore": tscore,
            "p": pval,
            "2.5%": lci,
            "97.5%": uci,
        },
        names,
    )
    aic = -2 * (-model.fun) + 2 * (len(coef))
    results = OpModel(
        model.fun, coef, aic, model.hess_inv, datasetnew, xs, ts, x_, yx_,
        yncat, x, y
    )
    return results


def opmod(data, x, y, pstart=None, method="BFGS", weights=1, offsetx=0):
    """Estimates Ordered Probit model and returns :class:`OpModel` class object.

    :param pstart: A list of starting values for the estimation. 
        Length of the number of parameters to be estimated.  
    :type pstart: list
    :param data: Full dataset used for estimation.
    :type data: pandas.DataFrame
    :type x: list of str
    :param y: Dependent Variable (DV).
    :type y: list of str
    :param method: Method for optimization, default is 'BFGS'. For other
        available methods, see scipy.optimize.minimize documentation.
    :param weights: Weights.
    :param offsetx: Offset for covariates (X).
    :return: OpModel
    """
    varlist = np.unique(y + x)
    dataset = data[varlist]
    datasetnew = dataset.dropna(how="any")
    x_ = datasetnew[x]
    y_ = datasetnew[y]
    yx_ = y_.iloc[:, 0]
    yncat = len(np.unique(yx_))
    if pstart is None:
        pstart = np.repeat(0.01, ((yncat - 1) + len(x_.columns)))
    model = minimize(
        op,
        pstart,
        args=(x_, yx_, datasetnew, weights, offsetx),
        method=method,
        options={"disp": True, "maxiter": 500},
    )
    results = opresults(model, data, x, y)
    return results


def iopresults(model, data, x, y, z, modeltype):
    """Produce estimation results, part of :py:func:`iopmod`.

    :param model: Model estimation results obtained from minimization.
    :param data: Full dataset used for estimation.
    :type data: pandas.DataFrame
    :param x: List of names for covariates in the ordered probit stage.
    :type x: list of str
    :param y: : List of names for dependent cariable.
    :type y: list of str
    :param z: : List of names for covariates in the inflation stage.
    :type z: list of str
    :param modeltype: : Type of model. Options are: 'ziop' or 'miop'.
    """
    varlist = np.unique(y + z + x)
    dataset = data[varlist]
    datasetnew = dataset.dropna(how="any")
    datasetnew = datasetnew.reset_index(drop=True)
    x_ = datasetnew[x]
    y_ = datasetnew[y]
    yx_ = y_.iloc[:, 0]
    yncat = len(np.unique(yx_))
    z_ = datasetnew[z]
    z_.insert(0, "int", np.repeat(1, len(z_)))
    names = list()
    for s in range(1, yncat):
        names.append("cut" + str(s))
    for s in range(z_.shape[1]):
        names.append("Inflation: " + z_.columns[s])
    for s in range(x_.shape[1]):
        names.append("Ordered: " + x_.columns[s])
    ts = model.x[0: yncat - 1]
    zs = model.x[yncat - 1: (yncat + z_.shape[1] - 1)]
    xs = model.x[
         (yncat + z_.shape[1] - 1): (yncat + z_.shape[1] + x_.shape[1] - 1)]
    ses = np.sqrt(np.diag(model.hess_inv))
    tscore = model.x / ses
    pval = (1 - (norm.cdf(abs(tscore)))) * 2
    lci = model.x - 1.96 * ses
    uci = model.x + 1.96 * ses
    coef = pd.DataFrame(
        {
            "Coef": model.x,
            "SE": ses,
            "tscore": tscore,
            "p": pval,
            "2.5%": lci,
            "97.5%": uci,
        },
        names,
    )
    aic = -2 * (-model.fun) + 2 * (len(coef))
    results = IopModel(
        modeltype,
        model.fun,
        coef,
        aic,
        model.hess_inv,
        datasetnew,
        xs,
        zs,
        ts,
        x_,
        yx_,
        z_,
        yncat,
        x,
        y,
        z,
    )
    return results


def iopcresults(model, data, x, y, z, modeltype):
    """Produce estimation results, part of :py:func:`ziopc  mod`.

    :param model: Model estimation results obtained from minimization.
    :param data: Full dataset used for estimation.
    :type data: pandas.DataFrame
    :param x: List of names for covariates in the ordered probit stage.
    :type x: list of str
    :param y: : List of names for dependent cariable.
    :type y: list of str
    :param z: : List of names for covariates in the inflation stage.
    :type z: list of str
    :param modeltype: : Type of model. Options are: 'ziopc' or 'miopc'
    """
    varlist = np.unique(y + z + x)
    dataset = data[varlist]
    datasetnew = dataset.dropna(how="any")
    datasetnew = datasetnew.reset_index(drop=True)
    x_ = datasetnew[x]
    y_ = datasetnew[y]
    yx_ = y_.iloc[:, 0]
    yncat = len(np.unique(yx_))
    z_ = datasetnew[z]
    z_.insert(0, "int", np.repeat(1, len(z_)))
    names = list()
    for s in range(1, yncat):
        names.append("cut" + str(s))
    for s in range(z_.shape[1]):
        names.append("Inflation: " + z_.columns[s])
    for s in range(x_.shape[1]):
        names.append("Ordered: " + x_.columns[s])
    names.append("rho")
    ts = model.x[0: yncat - 1]
    zs = model.x[yncat - 1: (yncat + z_.shape[1] - 1)]
    xs = model.x[
         (yncat + z_.shape[1] - 1): (yncat + z_.shape[1] + x_.shape[1] - 1)]
    rho = model.x[-1]
    ses = np.sqrt(np.diag(model.hess_inv))
    tscore = model.x / ses
    pval = (1 - (norm.cdf(abs(tscore)))) * 2
    lci = model.x - 1.96 * ses
    uci = model.x + 1.96 * ses
    coef = pd.DataFrame(
        {
            "Coef": model.x,
            "SE": ses,
            "tscore": tscore,
            "p": pval,
            "2.5%": lci,
            "97.5%": uci,
        },
        names,
    )
    aic = -2 * (-model.fun) + 2 * (len(coef))
    results = IopCModel(
        modeltype,
        model.fun,
        coef,
        aic,
        model.hess_inv,
        datasetnew,
        xs,
        zs,
        ts,
        x_,
        yx_,
        z_,
        rho,
        yncat,
        x,
        y,
        z,
    )
    return results


def iopmod(
        modeltype,
        data,
        x,
        y,
        z,
        pstart=None,
        method="BFGS",
        weights=1,
        offsetx=0,
        offsetz=0,
):
    """Estimate ZiOP model and return :class:`IopModel` class object as output.

    :param pstart: A list of starting values for the estimation.
        Length of the number of parameters to be estimated.
    :type pstart: list
    :param data: Full dataset used for estimation.
    :type data: pandas.DataFrame
    :param x: Covariates in the ordered probit stage.
        Elements must match column names of ``data``.
    :type x: list of str
    :param y: Dependent variable (DV). Element must match column
        names of ``data``.
    :type y: list of str
    :param z: Inflation stage variable. Elements must match column
        names of ``data``.
    :type z: list of str
    :param modeltype: Type of model to be estimated.
        Options are: "ziop" or 'miop'.
    :param method: Method for optimization, default 'BFGS'.  For other
        available methods, see scipy.optimize.minimize documentation.
    :param weights: Weights.
    :param offsetx: Offset for ordered probit stage covariates (X).
    :param offsetz: Offset for inflation stage covariates (Z).
    :return: IopModel
    """
    types = ["ziop", "miop"]
    if modeltype in types:
        varlist = np.unique(y + z + x)
        dataset = data[varlist]
        datasetnew = dataset.dropna(how="any")
        x_ = datasetnew[x]
        y_ = datasetnew[y]
        yx_ = y_.iloc[:, 0]
        yncat = len(np.unique(yx_))
        z_ = datasetnew[z]
        z_.insert(0, "ones", np.repeat(1, len(z_)))
        if pstart is None:
            pstart = np.repeat(0.01, (
                    (yncat - 1) + len(x_.columns) + len(z_.columns)))
        if modeltype == "ziop":
            model = minimize(
                ziop,
                pstart,
                args=(x_, yx_, z_, datasetnew, weights, offsetx, offsetz),
                method=method,
                options={"disp": True, "maxiter": 500},
            )
        elif modeltype == "miop":
            if len(np.unique(y_.astype("category").iloc[:, 0])) % 2 == 1:
                model = minimize(
                    miop,
                    pstart,
                    args=(x_, yx_, z_, datasetnew, weights, offsetx, offsetz),
                    method=method,
                    options={"disp": True, "maxiter": 500},
                )
            else:
                raise Exception("miop requires odd number of categories.")
        results = iopresults(model, data, x, y, z, modeltype)
        return results
    else:
        raise Exception("type must be ziop or miop")


def iopcmod(
        modeltype,
        data,
        x,
        y,
        z,
        pstart=None,
        method="BFGS",
        weights=1,
        offsetx=0,
        offsetz=0,
):
    """Estimate an iOP model (ZiOP or MiOP) and return :class:`IopcModel`.

    :param pstart: A list of starting values for the estimation.
        Length of the number of parameters to be estimated.
    :type pstart: list
    :param data: Dataset used for estimation.
    :type data: pandas.DataFrame
    :param x: Covariates for the ordered probit stage.
        Elements must match column names of ``data``.
    :type x: list of str
    :param y: The dependent variable (DV). Element must match column
        names of ``data``.
    :type y: list of str
    :param z: Covariates for the inflation stage. Elements must match column
        names of ``data``.
    :type z: list of str
    :param modeltype: Type of model to be estimated.
        Options are: 'ziopc' or 'miopc'.
    :param method: Method for optimization, default is 'BFGS'. For other
        available methods, see scipy.optimize.minimize documentation.
    :param weights: Weights.
    :param offsetx: Offset for variables in ordered probit stage (X).
    :param offsetz: Offset for variables in inflation stage (Z).
    :return: IopCModel
    """
    types = ["ziopc", "miopc"]
    if modeltype in types:
        varlist = np.unique(y + z + x)
        dataset = data[varlist]
        datasetnew = dataset.dropna(how="any")
        datasetnew = datasetnew.reset_index(drop=True)
        x_ = datasetnew[x]
        y_ = datasetnew[y]
        yx_ = y_.iloc[:, 0]
        yncat = len(np.unique(yx_))
        z_ = datasetnew[z]
        z_.insert(0, "ones", np.repeat(1, len(z_)))
        if pstart is None:
            pstart = np.repeat(
                0.01, ((yncat - 1) + len(x_.columns) + len(z_.columns) + 1)
            )
        if modeltype == "ziopc":
            model = minimize(
                ziopc,
                pstart,
                args=(x_, yx_, z_, datasetnew, weights, offsetx, offsetz),
                method=method,
                options={"disp": True, "maxiter": 500},
            )
        elif modeltype == "miopc":
            if len(np.unique(y_.astype("category").iloc[:, 0])) % 2 == 1:
                model = minimize(
                    miopc,
                    pstart,
                    args=(x_, yx_, z_, datasetnew, weights, offsetx, offsetz),
                    method=method,
                    options={"disp": True, "maxiter": 500},
                )
            else:
                raise Exception("miopc requires odd number of categories.")
                return
        results = iopcresults(model, data, x, y, z, modeltype)
        return results
    else:
        raise Exception("type must be ziopc or miopc")


def iopfit(model):
    """Calculate probabilities from :py:func:`iopmod`.

    :param model: :class:IopModel object from :py:func:`iopmod`.
    :return: :class:FittedVals object with fitted values.
    """
    zg = model.Z.dot(model.inflate)
    xb = model.X.dot(model.ordered)
    cprobs = np.zeros((model.ycat - 1, 1))
    n = len(model.data)
    probs = np.zeros((n, model.ycat))
    cprobs[0, 0] = model.cutpoints[0]
    if model.modeltype == "ziop":
        for j in range(1, model.ycat - 1):
            cprobs[j, 0] = cprobs[j - 1, 0] + np.exp(model.cutpoints[j])
        probs[:, model.ycat - 1] = (norm.cdf(zg)) * (
                1 - norm.cdf(cprobs[model.ycat - 2, 0] - xb)
        )
        probs[:, 0] = (1 - norm.cdf(zg)) + (norm.cdf(zg)) * (
            norm.cdf(cprobs[0, 0] - xb)
        )
        for j in range(1, model.ycat - 1):
            probs[:, j] = (norm.cdf(zg)) * ((norm.cdf(cprobs[j, 0] - xb))
                                            - (norm.cdf(cprobs[j - 1, 0]
                                                        - xb)))
    elif model.modeltype == "miop":
        for j in range(1, model.ycat - 1):
            cprobs[j, 0] = cprobs[j - 1, 0] + np.exp(model.cutpoints[j])
        probs[:, model.ycat - 1] = (norm.cdf(zg)) * (
                1 - norm.cdf(cprobs[model.ycat - 2, 0] - xb)
        )
        probs[:, 0] = norm.cdf(zg) * norm.cdf(cprobs[0, 0] - xb)

        for i in range(1, model.ycat - 1):
            if i == np.median(range(model.ycat)):
                probs[:, i] = (1 - norm.cdf(zg)) + (
                        norm.cdf(zg)
                        * (norm.cdf(cprobs[j, 0] - xb)
                           - norm.cdf(cprobs[j - 1, 0] - xb)))
            else:
                probs[:, i] = norm.cdf(zg) * (
                        norm.cdf(cprobs[:, i] - xb)
                        - norm.cdf(cprobs[j - 1, 0] - xb))
    probsordered = np.zeros((n, model.ycat))
    probsordered[:, model.ycat - 1] = 1 - norm.cdf(
        cprobs[model.ycat - 2, 0] - xb)
    probsordered[:, 0] = norm.cdf(cprobs[0, 0] - xb)
    for j in range(1, model.ycat - 1):
        probsordered[:, j] = (norm.cdf(cprobs[j, 0] - xb)) - (
            norm.cdf(cprobs[j - 1, 0] - xb)
        )
    probsinfl = np.zeros((n, 1))
    probsinfl[:, 0] = 1 - norm.cdf(zg)
    probslin = pd.DataFrame({"ZG": zg, "XB": xb})
    fitted = FittedVals(probs, probsordered, probsinfl, probslin)
    return fitted


def iopcfit(model):
    """Calculate fitted probabilities from :py:func:`iopcmod`.

    :param model: :class:`IopCModel` object from :py:func:`iopcmod`.
    :return: :class:`FittedVals` object with fitted values.
    """
    zg = model.Z.dot(model.inflate)
    xb = model.X.dot(model.ordered)
    cprobs = np.zeros((model.ycat - 1, 1))
    n = len(model.data)
    probs = np.zeros((n, model.ycat))
    cprobs[0, 0] = model.cutpoints[0]
    rho = model.coefs.iloc[-1, 0]
    means = np.array([0, 0])
    lower = np.array([-inf, -inf])
    sigma = np.array([[1, rho], [rho, 1]])
    nsigma = np.array([[1, -rho], [-rho, 1]])
    if model.modeltype == "ziopc":
        for j in range(1, model.ycat - 1):
            cprobs[j, 0] = cprobs[j - 1, 0] + np.exp(model.cutpoints[j])
        for i in range(n):
            probs[i, model.ycat - 1] = mvn.mvnun(
                lower, [zg[i], (xb[i] - cprobs[model.ycat - 2][0])], means,
                sigma
            )[0]
            probs[i, 0] = (1 - norm.cdf(zg[i])) + mvn.mvnun(
                lower, [zg[i], (cprobs[0][0] - xb[i])], means, nsigma
            )[0]
        for i in range(n):
            for j in range(1, model.ycat - 1):
                probs[i, j] = (
                                  mvn.mvnun(lower,
                                            [zg[i], (cprobs[j][0] - xb[i])],
                                            means, nsigma)[0]
                              ) - (
                                  mvn.mvnun(
                                      lower,
                                      [zg[i], (cprobs[j - 1][0] - xb[i])],
                                      means, nsigma
                                  )[0]
                              )
    elif model.modeltype == "miopc":
        for j in range(1, model.ycat - 1):
            cprobs[j, 0] = cprobs[j - 1, 0] + np.exp(model.cutpoints[j])
        for i in range(n):
            probs[i, model.ycat - 1] = mvn.mvnun(
                lower, [zg[i], (xb[i] - cprobs[model.ycat - 2][0])], means,
                sigma
            )[0]
            probs[i, 0] = mvn.mvnun(
                lower, [zg[i], (cprobs[0][0] - xb[i])], means, nsigma
            )[0]
        for i in range(n):
            for j in range(1, model.ycat - 1):
                if j == np.median(range(model.ycat)):
                    probs[i, j] = (1 - norm.cdf(zg[i])) + (
                            (mvn.mvnun(
                                lower, [zg[i], (cprobs[j][0] - xb[i])],
                                means, nsigma)[0])
                            - (mvn.mvnun(lower,
                                         [zg[i], (cprobs[j - 1][0] -
                                                  xb[i])],
                                         means, nsigma)[0]))
                else:
                    probs[i, j] = (mvn.mvnun(lower,
                                             [zg[i], (cprobs[j][0] - xb[i])],
                                             means, nsigma)[0]) - (
                                      mvn.mvnun(lower, [zg[i],
                                                        (cprobs[j - 1][0] -
                                                         xb[i])],
                                                means, nsigma)[0])

    # ordered
    probsordered = np.zeros((n, model.ycat))
    probsordered[:, model.ycat - 1] = 1 - norm.cdf(
        cprobs[model.ycat - 2, 0] - xb)
    probsordered[:, 0] = norm.cdf(cprobs[0, 0] - xb)
    for j in range(1, model.ycat - 1):
        probsordered[:, j] = norm.cdf(cprobs[j, 0] - xb) - (
            norm.cdf(cprobs[j - 1, 0] - xb)
        )

    probsinfl = np.zeros((n, 1))
    probsinfl[:, 0] = 1 - norm.cdf(zg)
    probslin = pd.DataFrame({"zg": zg, "xb": xb})
    fitted = FittedVals(probs, probsordered, probsinfl, probslin)
    return fitted


def vuong_opiop(opmodel, iopmodel):
    """Run the Vuong test to compare the performance of the OP and iOP model.

    :param opmodel: The OP model from :class:`OpModel`.
    :param iopmodel: The ZiOP model from :class:`IopModel`.
    :return: vuongopiop: Result of the Vuong test.
    """
    n1 = len(opmodel.data)
    y = iopmodel.Y
    # can also y = opmodel.Y when 2 models have the same length
    x = iopmodel.X
    # can also x = opmodel.X when 2 models have the same length
    cuts_op = np.repeat(0, len(opmodel.cutpoints)).astype(float)
    xop = opmodel.ordered
    cuts_op[0] = opmodel.cutpoints[0]
    for i in range(1, len(opmodel.cutpoints)):
        cuts_op[i] = cuts_op[i - 1] + np.exp(opmodel.cutpoints[i])
    xbop = pd.DataFrame(index=np.arange(n1), columns=np.arange(len(x.columns)))
    for j in range(len(x.columns)):
        xbop.iloc[:, j] = xop[j] * x.iloc[:, j]
    xbop_sum = xbop.sum(axis=1)
    fitttediop = iopfit(iopmodel).responsefull
    ycat = y.astype("category")
    ycatu = np.unique(ycat)
    yncat = len(ycatu)
    y0 = np.sort(ycatu)
    v = np.zeros((n1, yncat))
    for j in range(yncat):
        v[:, j] = y == y0[j]
    m = np.zeros(n1)
    probs = np.zeros((n1, yncat))
    probs[:, 0] = norm.cdf(cuts_op[0] - xbop_sum) / fitttediop[:, 0]
    probs[:, yncat - 1] = (1 - norm.cdf(
        cuts_op[yncat - 2] - xbop_sum)) / fitttediop[:, yncat - 1]
    for i in range(1, yncat - 1):
        probs[:, i] = (norm.cdf(cuts_op[i] - xbop_sum)
                       - norm.cdf(cuts_op[i - 1]
                                  - xbop_sum)) / fitttediop[:, i]
    m = np.zeros((n1, yncat))
    for k in range(n1):
        for j in range(yncat):
            m[k, j] = v[k, j] * probs[k, j]
    m2 = m[m != 0]
    mlog = np.log(m2)
    diffmsq = (mlog - np.mean(mlog)) ** 2
    sumdms = sum(diffmsq)
    vuongopiop = (np.sqrt(n1) * (1 / n1) * sum(mlog)) / (
        np.sqrt((1 / n1) * sumdms))
    return vuongopiop


def vuong_opiopc(opmodel, iopcmodel):
    """Run the Vuong test to compare the performance of the OP and iOPC model.

    :param opmodel: The OP model from :class:`OpModel`.
    :param iopcmodel: The iOPC model from :class:`IopCModel`.
    :return: vuongopiopc: Result of the Vuong test.
    """
    n1 = len(opmodel.data)
    y = iopcmodel.Y
    x = iopcmodel.X
    cuts_op = np.repeat(0, len(opmodel.cutpoints)).astype(float)
    xop = opmodel.ordered
    cuts_op[0] = opmodel.cutpoints[0]
    for i in range(1, len(opmodel.cutpoints)):
        cuts_op[i] = cuts_op[i - 1] + np.exp(opmodel.cutpoints[i])
    xbop = pd.DataFrame(index=np.arange(n1), columns=np.arange(len(x.columns)))
    for j in range(len(x.columns)):
        xbop.iloc[:, j] = xop[j] * x.iloc[:, j]
    xbop_sum = xbop.sum(axis=1)
    fitttedziopc = iopcfit(iopcmodel).responsefull
    ycat = y.astype("category")
    ycatu = np.unique(ycat)
    yncat = len(ycatu)
    y0 = np.sort(ycatu)
    v = np.zeros((n1, yncat))
    for j in range(yncat):
        v[:, j] = y == y0[j]
    probs = np.zeros((n1, yncat))
    probs[:, 0] = norm.cdf(cuts_op[0] - xbop_sum) / fitttedziopc[:, 0]
    probs[:, yncat - 1] = (1 - norm.cdf(
        cuts_op[yncat - 2] - xbop_sum)) / fitttedziopc[:, yncat - 1]
    for i in range(1, yncat - 1):
        probs[:, i] = (norm.cdf(cuts_op[i] - xbop_sum) - norm.cdf(
            cuts_op[i - 1] - xbop_sum)) / fitttedziopc[:, i]
    m = np.zeros((n1, yncat))
    for k in range(n1):
        for j in range(yncat):
            m[k, j] = v[k, j] * probs[k, j]
    m2 = m[m != 0]
    mlog = np.log(m2)
    diffmsq = (mlog - np.mean(mlog)) ** 2
    sumdms = sum(diffmsq)
    vuongopiopc = (np.sqrt(n1) * (1 / n1) * sum(mlog)) / (
        np.sqrt((1 / n1) * sumdms))
    return vuongopiopc


def split_effects(model, inflvar, nsims=10000):
    """Calculate change in probability of being 0 in the split-probit stage.

    This function calculates the predicted probabilities
    when there is change in value of a variable in the split-probit equation.
    The chosen dummy variable is changed from 0 to 1,
    and chosen numerical variable is mean value + 1 standard deviation.
    Other variables are kept at 0 or mean value
    (Note: the current version of the function
    recognizes ordinal variables as numerical).

    :param model: :class:`IopModel` or :class:`IopCModel`.
    :param inflvar: Number representing the location of variable
        in the split-probit equation.
        (attribute .inflate of :class:`IopModel` or :class:`IopCModel`)
    :type inflvar: int
    :param nsims: Number of simulated observations, default is 10000.
    :type nsims: int
    :return: changeprobs: A dataframe of the predicted
        probabilities when there is change in the variable (1)
        versus original values (0).
    """
    estimate = model.coefs.iloc[:, 0]
    vcov = model.vcov
    model_z = model.Z
    zsim1 = np.zeros(len(model_z.columns))
    zsim1[0] = 1
    zsima = np.zeros(len(model_z.columns))
    zsima[0] = 1
    for j in range(1, len(model_z.columns)):
        if (
                max(model_z.iloc[:, j]) == 1
                and min(model_z.iloc[:, j]) == 0
                and len(np.unique(model_z.iloc[:, j])) == 2
        ):
            zsim1[j] = 0
        else:
            zsim1[j] = np.mean(model_z.iloc[:, j])
    for j in range(1, len(model_z.columns)):
        if (
                max(model_z.iloc[:, j]) == 1
                and min(model_z.iloc[:, j]) == 0
                and len(np.unique(model_z.iloc[:, j])) == 2
        ):
            zsima[j] = 1
        else:
            zsima[j] = np.mean(model_z.iloc[:, j]) + np.std(model_z.iloc[:, j])
    zsim2 = zsim1.copy()
    zsim2[inflvar] = zsima[inflvar]
    np.random.seed(1)
    probs1 = np.zeros(nsims)
    probs2 = np.zeros(nsims)
    for i in range(nsims):
        gsim = np.random.multivariate_normal(estimate, vcov)
        gsim2 = gsim[model.ycat - 1: model.ycat - 1 + len(model.inflate)]
        zg1 = zsim1.dot(gsim2)
        zg2 = zsim2.dot(gsim2)
        probs1[i] = norm.cdf(zg1)
        probs2[i] = norm.cdf(zg2)
    name = model.coefs.index[model.ycat - 1 + inflvar]
    changeprobs = pd.DataFrame({name.replace("Inflation: ", "") + "= 0": probs1,
                                name.replace("Inflation: ", "") + "= 1": probs2}
                               )
    return changeprobs


def ordered_effects(model, ordvar, nsims=10000):
    """Calculate the changes in probability in each outcome in OP stage.

    This function calculates predicted probabilities
    when there is a change in the value of a covariate
    in the ordered probit equation.
    The chosen dummy variable is changed from 0 to 1,
    and chosen numerical variable is mean value + 1 standard deviation.
    Other variables are kept at 0 or mean value
    (Note: the current version of the function
    recognizes ordinal variables as numerical).

    :param model: :class:`IopModel` or :class:`IopCModel`.
    :param ordvar: Number representing the location of variable
        in the ordered probit equation.
        (attribute .ordered of :class:`IopModel` or :class:`IopCModel`)
    :type ordvar: int
    :param nsims: Number of simulated observations, default is 10000.
    :type nsims: int
    :return: changeprobs: A dataframe of the predicted
        probabilities when there is change in the variable for each outcome (1)
        versus original values (0).
    """
    estimate = model.coefs.iloc[:, 0]
    vcov = model.vcov
    model_x = model.X
    xsim1 = np.zeros(len(model_x.columns))
    xsima = np.zeros(len(model_x.columns))
    for j in range(len(model_x.columns)):
        if (
                max(model_x.iloc[:, j]) == 1
                and min(model_x.iloc[:, j]) == 0
                and len(np.unique(model_x.iloc[:, j])) == 2
        ):
            xsim1[j] = 0
        else:
            xsim1[j] = np.mean(model_x.iloc[:, j])
    for j in range(len(model_x.columns)):
        if (
                max(model_x.iloc[:, j]) == 1
                and min(model_x.iloc[:, j]) == 0
                and len(np.unique(model_x.iloc[:, j])) == 2
        ):
            xsima[j] = 1
        else:
            xsima[j] = np.mean(model_x.iloc[:, j]) + np.std(model_x.iloc[:, j])
    xsim2 = xsim1.copy()
    xsim2[ordvar] = xsima[ordvar]
    np.random.seed(1)
    probsordered1 = np.zeros(model.ycat)
    probsordered2 = np.zeros(model.ycat)
    cprobs = np.zeros((model.ycat - 1, 1))
    cprobs[0, 0] = model.cutpoints[0]
    for j in range(1, model.ycat - 1):
        cprobs[j, 0] = cprobs[j - 1, 0] + np.exp(model.cutpoints[j])
    probs1 = pd.DataFrame(index=np.arange(nsims),
                          columns=np.arange(model.ycat))
    probs2 = pd.DataFrame(index=np.arange(nsims),
                          columns=np.arange(model.ycat))
    name = model.coefs.index[model.ycat - 1 + len(model.inflate) + ordvar]
    probs1 = probs1.add_suffix(": " + name.replace("Ordered: ", "") + " = 0")
    probs2 = probs2.add_suffix(": " + name.replace("Ordered: ", "") + " = 1")
    for i in range(nsims):
        bsim = np.random.multivariate_normal(estimate, vcov)
        bsim2 = bsim[model.ycat - 1 + len(model.inflate):
                     model.ycat - 1 + len(model.inflate) + len(model.ordered)]
        xb1 = xsim1.dot(bsim2)
        xb2 = xsim2.dot(bsim2)
        probsordered1[model.ycat - 1] = 1 - norm.cdf(
            cprobs[model.ycat - 2, 0] - xb1)
        probsordered1[0] = norm.cdf(cprobs[0, 0] - xb1)
        for j in range(1, model.ycat - 1):
            probsordered1[j] = norm.cdf(cprobs[j, 0] - xb1) - (
                norm.cdf(cprobs[j - 1, 0] - xb1)
            )
        probsordered2[model.ycat - 1] = 1 - norm.cdf(
            cprobs[model.ycat - 2, 0] - xb2)
        probsordered2[0] = norm.cdf(cprobs[0, 0] - xb2)
        for j in range(1, model.ycat - 1):
            probsordered2[j] = norm.cdf(cprobs[j, 0] - xb2) - (
                norm.cdf(cprobs[j - 1, 0] - xb2)
            )
        probs1.iloc[i:, ] = probsordered1
        probs2.iloc[i:, ] = probsordered2
    changeprobs = pd.DataFrame(index=np.arange(nsims),
                               columns=np.arange(2 * model.ycat))
    newnames = list(np.repeat("", model.ycat * 2))
    for j in range(0, 2 * model.ycat, 2):
        changeprobs.iloc[:, j] = probs1.iloc[:, round(j / 2)]
    for j in range(1, 2 * model.ycat, 2):
        changeprobs.iloc[:, j] = probs2.iloc[:, round((j - 1) / 2)]
    for j in range(0, 2 * model.ycat, 2):
        newnames[j] = list(probs1.columns)[round(j / 2)]
    for j in range(1, 2 * model.ycat, 2):
        newnames[j] = list(probs2.columns)[round((j - 1) / 2)]
    changeprobs.columns = newnames
    return changeprobs



"""Classes and Functions for the gimnl module."""
import numpy as np
import pandas as pd
from scipy.optimize import minimize
from scipy.stats import norm


class GimnlModel:
    """Store model results from :py:func:`gimnlmod`."""

    def __init__(
            self,
            modeltype,
            reference,
            inflatecat,
            llik,
            coef,
            aic,
            vcov,
            data,
            xs,
            zs,
            x_,
            yx_,
            z_,
            ycatu,
            xstr,
            ystr,
            zstr,
    ):
        """Store model results, goodness-of-fit tests, and other information
        for the Generalized Inflated Multinomial Logit Model.

        :param modeltype: Type of GIMNL Model (bimnl3, simnl3, or timnl3),
            indicating the inflated category.
        :param reference: Order of categories. The order category will be
            the first element.
        :param llik: Log-Likelihood.
        :param coef: Model coefficients.
        :param aic: Model Akaike information .
        :param vcov: Variance-Covariance matrix.
            (optimized as inverted Hessian matrix)
        :param data: Model data used for estimation, subsetted to selected
            variables and missing values listwise deleted.
        :param zs: Inflation stage estimates (Gammas).
        :param xs: Multinomial Logit probit estimates (Betas).
        :param ycatu: Number of categories in the Dependent variable (DV).
        :param x_: X Data.
        :param yx_: Y (DV) data.
        :param z_: Z Data.
        :param xstr: List of string for variable names in the MNL stage.
        :param ystr: List of string for dependent variable name .
        :param zstr: List of string for variable names in the Logit Split
        stage.
        """
        self.modeltype = modeltype
        self.reference = reference
        self.inflatecat = inflatecat
        self.llik = llik
        self.coefs = coef
        self.AIC = aic
        self.vcov = vcov
        self.data = data
        self.split = zs
        self.multinom = xs
        self.ycat = ycatu
        self.X = x_
        self.Y = yx_
        self.Z = z_
        self.xstr = xstr
        self.ystr = ystr
        self.zstr = zstr


class MnlModel:
    """Store model results from :py:func:`gimnlmod`."""

    def __init__(
            self,
            modeltype,
            reference,
            llik,
            coef,
            aic,
            vcov,
            data,
            xs,
            x_,
            yx_,
            ycatu,
            xstr,
            ystr,
    ):
        """Store model results, goodness-of-fit tests, and other information
        for the Multinomial Logit Model.

        :param modeltype: Type of Model (mnl3).
        :param reference: Order of categories. The order category will be
            the first element.
        :param llik: Log-Likelihood.
        :param coef: Model coefficients.
        :param aic: Model Akaike information .
        :param vcov: Variance-Covariance matrix.
            (optimized as inverted Hessian matrix)
        :param data: Model data used for estimation, subsetted to selected
            variables and missing values listwise deleted.
        :param xs: Multinomial Logit estimates (Betas).
        :param ycatu: Number of categories in the Dependent variable (DV).
        :param x_: X Data.
        :param yx_: Y (DV) data.
        :param xstr: List of string for x names.
        :param ystr: List of string for y names.
        """
        self.modeltype = modeltype
        self.reference = reference
        self.llik = llik
        self.coefs = coef
        self.AIC = aic
        self.vcov = vcov
        self.data = data
        self.multinom = xs
        self.ycat = ycatu
        self.X = x_
        self.Y = yx_
        self.xstr = xstr
        self.ystr = ystr


def mnl3(pstart, x2, x3, y, reference):
    """
    Calculate likelihood function for the baseline inflated three-category MNL
    model.

    :param pstart: A list of starting parameters.
    :type pstart: list
    :param x2: Multinomial Logit variables. Data subsetted to selected
        variables.
    :type x2: Pandas dataframe
    :param x3: Multinomial Logit variables (should be identical to x2.
    :type x3: Pandas dataframe
    :param y: Dependent variable (DV). Data subsetted to selected
        variable.
    :type y: Pandas dataframe
    :param reference: List of order of categories.
    """
    b2 = pstart[0: len(x2.columns)]
    b3 = pstart[len(x2.columns): (len(pstart))]
    xb2 = x2.dot(b2)
    xb3 = x3.dot(b3)
    p1 = 1 / (1 + np.exp(xb2) + np.exp(xb3))
    p2 = p1 * np.exp(xb2)
    p3 = p1 * np.exp(xb3)
    lik = np.sum(
        np.log(p1) * (y == reference[0])
        + np.log(p2) * (y == reference[1])
        + np.log(p3) * (y == reference[2])
    )
    llik = -1 * np.sum(lik)
    return llik


def bimnl3(pstart, x2, x3, y, z, reference):
    """
    Calculate likelihood function for the baseline inflated three-category
    MNL model.

    :param pstart: A list of starting parameters.
    :type pstart: list
    :param x2: Multinomial Logit variables. Data subsetted to selected
        variables.
    :type x2: pandas dataframe
    :param x3: Multinomial Logit variables (should be identical to x2.
    :type x3: pandas dataframe
    :param y: Dependent variable (DV). Data subsetted to selected
        variable.
    :type y: pandas dataframe
    :param z: Logit Split stage variables. Data subsetted to selected
        variables.
    :type z: pandas dataframe
    :param reference: List of order of categories (first element will be the
        inflated category).
    """
    b2 = pstart[len(z.columns): (len(z.columns) + len(x2.columns))]
    b3 = pstart[(len(z.columns) + len(x2.columns)): (len(pstart))]
    gamma = pstart[0: (len(z.columns))]
    xb2 = x2.dot(b2)
    xb3 = x3.dot(b3)
    zg = z.dot(gamma)
    pz = 1 / (1 + np.exp(-zg))
    p1 = 1 / (1 + np.exp(xb2) + np.exp(xb3))
    p2 = p1 * np.exp(xb2)
    p3 = p1 * np.exp(xb3)
    lik = np.sum(
        np.log((1 - pz) + pz * p1) * (y == reference[0])
        + np.log(pz * p2) * (y == reference[1])
        + np.log(pz * p3) * (y == reference[2])
    )
    llik = -1 * np.sum(lik)
    return llik


def simnl3(pstart, x2, x3, y, z, reference):
    """
    Calculate likelihood function for the second category inflated
    three-category MNL model.

    :param pstart: A list of starting parameters.
    :type pstart: list
    :param x2: Multinomial Logit variables. Data subsetted to selected
        variables.
    :type x2: pandas dataframe
    :param x3: Multinomial Logit variables (should be identical to x2.
    :type x3: pandas dataframe
    :param y: Dependent variable (DV). Data subsetted to selected
        variable.
    :type y: pandas dataframe
    :param z: Logit Split stage variables. Data subsetted to selected
        variables.
    :type z: pandas dataframe
    :param reference: List of order of categories (second element will be the
        inflated category).
    """
    b2 = pstart[len(z.columns): (len(z.columns) + len(x2.columns))]
    b3 = pstart[(len(z.columns) + len(x2.columns)): (len(pstart))]
    gamma = pstart[0: (len(z.columns))]
    xb2 = x2.dot(b2)
    xb3 = x3.dot(b3)
    zg = z.dot(gamma)
    pz = 1 / (1 + np.exp(-zg))
    p1 = 1 / (1 + np.exp(xb2) + np.exp(xb3))
    p2 = p1 * np.exp(xb2)
    p3 = p1 * np.exp(xb3)
    lik = np.sum(
        np.log(pz * p1) * (y == reference[0])
        + np.log((1 - pz) + pz * p2) * (y == reference[1])
        + np.log(pz * p3) * (y == reference[2])
    )
    llik = -1 * np.sum(lik)
    return llik


def timnl3(pstart, x2, x3, y, z, reference):
    """
    Calculate likelihood function for the third category inflated
    three-category MNL model.

    :param pstart: A list of starting parameters.
    :type pstart: list
    :param x2: Covariates in the MNL. 
    :type x2: pandas dataframe
    :param x3: Multinomial Logit variables (should be identical to x2).
    :type x3: pandas dataframe
    :param y: Dependent variable (DV). 
    :type y: pandas dataframe
    :param z: Covariates in the Logit model (i.e. split stage). 
        variables.
    :type z: pandas dataframe
    :param reference: List of order of categories (third element will be the
        inflated category).
    """
    b2 = pstart[len(z.columns): (len(z.columns) + len(x2.columns))]
    b3 = pstart[(len(z.columns) + len(x2.columns)): (len(pstart))]
    gamma = pstart[0: (len(z.columns))]
    xb2 = x2.dot(b2)
    xb3 = x3.dot(b3)
    zg = z.dot(gamma)
    pz = 1 / (1 + np.exp(-zg))
    p1 = 1 / (1 + np.exp(xb2) + np.exp(xb3))
    p2 = p1 * np.exp(xb2)
    p3 = p1 * np.exp(xb3)
    lik = np.sum(
        np.log(pz * p1) * (y == reference[0])
        + np.log(pz * p2) * (y == reference[1])
        + np.log((1 - pz) + pz * p3) * (y == reference[2])
    )
    llik = -1 * np.sum(lik)
    return llik


def gimnlresults(model, data, x, y, z, modeltype, reference, inflatecat):
    """
    Produce estimation results, part of :py:func:`gimnlmod`.

    Store estimates, model AIC, and other information to
    :py:class:`GimnlModel`.

    :param model: Estimation results.
    :param data: Model data used for estimation, subsetted to selected
        variables and missing values listwise deleted.
    :type data: pandas.DataFrame
    :param x: List of names for covariates in Multinomial Logit stage.
    :type x: list of str
    :param y: List of names for dependent variable .
    :type y: list of str
    :param z: List of names for covariates in split-stage.
    :type z: list of str
    :param modeltype: Type of inflated MNL model. One of 'bimnl3', 'simnl3',
        or 'timnl3'.
    :param reference: List of order of categories. First element is the
        baseline/ reference category.
    :param inflatecat: Inflated category. One of 'baseline', 'second',
        or 'third'.
    """
    varlist = np.unique(y + z + x)
    dataset = data[varlist]
    datasetnew = dataset.dropna(how="any")
    datasetnew = datasetnew.reset_index(drop=True)
    x_ = datasetnew[x]
    y_ = datasetnew[y]
    yx_ = y_.iloc[:, 0]
    yncat = len(np.unique(yx_))
    z_ = datasetnew[z]
    z_.insert(0, "int", np.repeat(1, len(z_)))
    x_.insert(0, "int", np.repeat(1, len(x_)))
    names = list()
    if modeltype == "bimnl3":
        x2 = x_
        x3 = x_
        for s in range(z_.shape[1]):
            names.append("Inflation: " + z_.columns[s])
        for s in range(x2.shape[1]):
            names.append(str(reference[1]) + ": " + x2.columns[s])
        for s in range(x3.shape[1]):
            names.append(str(reference[2]) + ": " + x3.columns[s])
        xs = model.x[(z_.shape[1]): (z_.shape[1] + x2.shape[1] + x3.shape[1])]
    zs = model.x[0: (z_.shape[1])]
    ses = np.sqrt(np.diag(model.hess_inv))
    tscore = model.x / ses
    pval = (1 - (norm.cdf(abs(tscore)))) * 2
    lci = model.x - 1.96 * ses
    uci = model.x + 1.96 * ses
    coef = pd.DataFrame(
        {
            "Coef": model.x,
            "SE": ses,
            "tscore": tscore,
            "p": pval,
            "2.5%": lci,
            "97.5%": uci,
        },
        names,
    )
    aic = -2 * (-model.fun) + 2 * (len(coef))
    llik = -1 * model.fun
    model = GimnlModel(
        modeltype,
        reference,
        inflatecat,
        llik,
        coef,
        aic,
        model.hess_inv,
        datasetnew,
        xs,
        zs,
        x_,
        yx_,
        z_,
        yncat,
        x,
        y,
        z,
    )
    return model


def mnlresults(model, data, x, y, modeltype, reference):
    """
    Produce estimation results, part of :py:func:`mnlmod`.

    Store estimates, model AIC, and other information to :py:class:`MnlModel`.

    :param model: Estimation results.
    :param data: Model data used for estimation, subsetted to selected
        variables and missing values listwise deleted.
    :type data: pandas.DataFrame
    :param x: List of names for covariates in Multinomial Logit stage.
    :type x: list of str
    :param y: List of names for dependent variable .
    :type y: list of str
    :param modeltype: Three-category MNL model ('mnl3').
    :param reference: List of order of categories. First element is the
        baseline/ reference category.
    """
    varlist = np.unique(y + x)
    dataset = data[varlist]
    datasetnew = dataset.dropna(how="any")
    datasetnew = datasetnew.reset_index(drop=True)
    x_ = datasetnew[x]
    y_ = datasetnew[y]
    yx_ = y_.iloc[:, 0]
    yncat = len(np.unique(yx_))
    x_.insert(0, "int", np.repeat(1, len(x_)))
    names = list()
    if modeltype == "mnl3":
        x2 = x_
        x3 = x_
        for s in range(x2.shape[1]):
            names.append(str(reference[1]) + ": " + x2.columns[s])
        for s in range(x3.shape[1]):
            names.append(str(reference[2]) + ": " + x3.columns[s])
        xs = model.x[0: (x2.shape[1] + x3.shape[1])]
    ses = np.sqrt(np.diag(model.hess_inv))
    tscore = model.x / ses
    pval = (1 - (norm.cdf(abs(tscore)))) * 2
    lci = model.x - 1.96 * ses
    uci = model.x + 1.96 * ses
    coef = pd.DataFrame(
        {
            "Coef": model.x,
            "SE": ses,
            "tscore": tscore,
            "p": pval,
            "2.5%": lci,
            "97.5%": uci,
        },
        names,
    )
    aic = -2 * (-model.fun) + 2 * (len(coef))
    llik = -1 * model.fun
    model = MnlModel(
        modeltype,
        reference,
        llik,
        coef,
        aic,
        model.hess_inv,
        datasetnew,
        xs,
        x_,
        yx_,
        yncat,
        x,
        y,
    )
    return model


def gimnlmod(data, x, y, z, reference, inflatecat, method="BFGS", pstart=None):
    """
    Estimate three-category inflated Multinomial Logit model.

    :param data: Full dataset.
    :type data: pandas.dataframe
    :param x: Covariates in Multi Nomial Logit stage. The elements must match
        column names of ``data``.
    :type x: list of str
    :param y: Dependent Variable. Values should be integers,
        with a number from 0-2 representing each category. The element
        must match column names of ``data``.
    :type y: list of str
    :param z: Covariates in split-stage. The elements must match column
        names of ``data``.
    :type z: list of str
    :param reference:  List of three elements specifying the order of
        categories (e.g [0, 1, 2], [2, 1, 0]. etc...). The first element is
        the baseline/reference category. The parameter
        inflatecat then specifies which category in the list inflated.
    :type reference: list of int
    :param inflatecat: A string specifying the inflated category. One of
        "baseline" for the first, "second" for the second, or "third" for the
        third in reference list to specify the inflated category.
    :param method: Optimization method.  Default is 'BFGS'. For other
        available methods, see scipy.optimize.minimize documentation.
    :param pstart: A list of starting parameters.
    :type pstart: list
    """
    varlist = np.unique(y + z + x)
    dataset = data[varlist]
    datasetnew = dataset.dropna(how="any")
    datasetnew = datasetnew.reset_index(drop=True)
    x_ = datasetnew[x]
    y_ = datasetnew[y]
    yx_ = y_.iloc[:, 0]
    yncat = len(np.unique(yx_))
    if yncat == 3:
        modeltype = "bimnl3"
    else:
        raise Exception(
            "Function only supports Dependent Variable with 3 " "categories."
        )
    z_ = datasetnew[z]
    z_.insert(0, "int", np.repeat(1, len(z_)))
    x_.insert(0, "int", np.repeat(1, len(x_)))
    if modeltype == "bimnl3":
        x2 = x_
        x3 = x_
        if pstart is None:
            pstart = np.repeat(
                0.01, (len(x2.columns) + len(x3.columns) + len(z_.columns))
            )
        if inflatecat == "baseline":
            model = minimize(
                bimnl3,
                pstart,
                args=(x2, x3, yx_, z_, reference),
                method=method,
                options={"disp": True, "maxiter": 500},
            )
        elif inflatecat == "second":
            model = minimize(
                simnl3,
                pstart,
                args=(x2, x3, yx_, z_, reference),
                method=method,
                options={"disp": True, "maxiter": 500},
            )
        elif inflatecat == "third":
            model = minimize(
                timnl3,
                pstart,
                args=(x2, x3, yx_, z_, reference),
                method=method,
                options={"disp": True, "maxiter": 500},
            )
    results = gimnlresults(model, data, x, y, z, modeltype, reference,
                           inflatecat)
    return results


def mnlmod(data, x, y, reference, method="BFGS", pstart=None):
    """
    Estimate three-category Multinomial Logit model.

    :param data: Full dataset.
    :type data: pandas.dataframe
    :param x: Covariates in MNL stage. The elements must match column
        names of ``data``
    :type x: list of str
    :param y: Dependent variable. Values should be integers,
        with a number from 0-2 representing each category.
    :param reference:  List of three elements specifying the order of
        categories (e.g [0, 1, 2], [2, 1, 0]. etc...). The first element is
        the baseline/reference category.
    :type reference: list of int
    :param method: Optimization method.  Default is 'BFGS'. For other
        available methods, see scipy.optimize.minimize documentation.
    :param pstart: A list of starting parameters.
    :type pstart: list
    """
    varlist = np.unique(y + x)
    dataset = data[varlist]
    datasetnew = dataset.dropna(how="any")
    datasetnew = datasetnew.reset_index(drop=True)
    x_ = datasetnew[x]
    y_ = datasetnew[y]
    yx_ = y_.iloc[:, 0]
    yncat = len(np.unique(yx_))
    if yncat == 3:
        modeltype = "mnl3"
    else:
        raise Exception(
            "Function only supports Dependent Variable with 3 " "categories."
        )
    x_.insert(0, "int", np.repeat(1, len(x_)))
    if modeltype == "mnl3":
        x2 = x_
        x3 = x_
        if pstart is None:
            pstart = np.repeat(
                0.01, (len(x2.columns) + len(x3.columns))
            )
            model = minimize(
                mnl3,
                pstart,
                args=(x2, x3, yx_, reference),
                method=method,
                options={"disp": True, "maxiter": 500},
            )
    results = mnlresults(model, data, x, y, modeltype, reference)
    return results


def vuong_gimnl(modelmnl, modelgimnl):
    """
    Run the Vuong test to compare the performance of the MNL and GIMNL model.

    For the function to run properly, the models need to have the same X
    covariates and same number of observations.

    :param modelmnl: A :class:`MnlModel` object.
    :param modelgimnl: A :class:`GimnlModel` object.
    """
    xb2_mnl = modelmnl.X.dot(modelmnl.multinom[0: len(modelmnl.X.columns)])
    xb3_mnl = modelmnl.X.dot(modelmnl.multinom[len(modelmnl.X.columns):
                                               len(modelmnl.multinom)])
    xb2_gimnl = modelgimnl.X.dot(
        modelgimnl.multinom[0: len(modelgimnl.X.columns)])
    xb3_gimnl = modelgimnl.X.dot(modelgimnl.multinom[len(modelgimnl.X.columns):
                                                     len(modelgimnl.multinom)])
    zg_gimnl = modelgimnl.Z.dot(modelgimnl.split)
    mnldenom = 1 + np.exp(xb2_mnl) + np.exp(xb3_mnl)
    gimnldenom = 1 + np.exp(xb2_gimnl) + np.exp(xb3_gimnl)
    p1mnl = 1 / mnldenom
    p2mnl = np.exp(xb2_mnl) / mnldenom
    p3mnl = np.exp(xb3_mnl) / mnldenom
    if modelgimnl.inflatecat == "baseline":
        p1gimnl = ((1 / gimnldenom)
                   * (1 / (1 + np.exp(-zg_gimnl)))
                   + (1 - (1 / (1 + np.exp(-zg_gimnl)))))
        p2gimnl = (np.exp(xb2_gimnl) / gimnldenom) * (
                1 / (1 + np.exp(-zg_gimnl)))
        p3gimnl = (np.exp(xb3_gimnl) / gimnldenom) * (
                1 / (1 + np.exp(-zg_gimnl)))
    elif modelgimnl.inflatecat == "second":
        p1gimnl = ((1 / gimnldenom)
                   * (1 / (1 + np.exp(-zg_gimnl))))
        p2gimnl = ((np.exp(xb2_gimnl) / gimnldenom)
                   * (1 / (1 + np.exp(-zg_gimnl)))
                   + (1 - (1 / (1 + np.exp(-zg_gimnl)))))
        p3gimnl = ((np.exp(xb3_gimnl) / gimnldenom)
                   * (1 / (1 + np.exp(-zg_gimnl))))
    elif modelgimnl.inflatecat == "third":
        p1gimnl = ((1 / gimnldenom)
                   * (1 / (1 + np.exp(-zg_gimnl))))
        p2gimnl = ((np.exp(xb2_gimnl) / gimnldenom)
                   * (1 / (1 + np.exp(-zg_gimnl))))
        p3gimnl = ((np.exp(xb3_gimnl) / gimnldenom)
                   * (1 / (1 + np.exp(-zg_gimnl)))
                   + (1 - (1 / (1 + np.exp(-zg_gimnl)))))
    m = np.zeros(len(modelgimnl.X))
    reference = modelgimnl.reference
    y = modelgimnl.Y
    for i in range(len(m)):
        if y[i] == reference[0]:
            m[i] = np.log(p1mnl[i] / p1gimnl[i])
        elif y[i] == reference[0]:
            m[i] = np.log(p2mnl[i] / p2gimnl[i])
        elif y[i] == reference[2]:
            m[i] = np.log(p3mnl[i] / p3gimnl[i])
    diffmsq = (m - np.mean(m)) ** 2
    sumdms = sum(diffmsq)
    vuong = (np.sqrt(len(m)) * (1 / len(m)) * sum(m)) / (
        np.sqrt((1 / len(m)) * sumdms))
    return vuong



import numpy as np
import pandas as pd
from idcempy import zmiopc
import os

DAT = pd.read_stata(os.getcwd()+"/data/bp_exact_for_analysis.dta")
X = ['logGDPpc', 'parliament', 'disaster', 'major_oil', 'major_primary']
Xsmall = ['logGDPpc', 'parliament', 'disaster']
Z = ['logGDPpc', 'parliament']
Y = ['rep_civwar_DV']
data = DAT

pstartziop = [-1.31, .32, 2.5, -.21, .2, -0.2, -0.4, 0.2, .9, -.4]

pstartziopc = [-1.31, .32, 2.5, -.21,
               .2, -0.2, -0.4, 0.2, .9, -.4, .1]


def test_ziopc_coefs():
    ziopc_JCR = zmiopc.iopcmod('ziopc',
                               data, X, Y, Z, pstart=pstartziopc)
    assert len(ziopc_JCR.coefs) == len(pstartziopc)


def test_ziop_coefs():
    ziop_JCR = zmiopc.iopmod('ziop', data, X, Y, Z, pstart=pstartziop)
    assert len(ziop_JCR.coefs) == len(pstartziop)


def test_iopc_fitted():
    ziopc_JCR = zmiopc.iopcmod('ziopc',
                               data, X, Y, Z, pstart=pstartziopc)
    fitttedziopc = zmiopc.iopcfit(ziopc_JCR)
    assert len(fitttedziopc.responseordered) == len(fitttedziopc.responsefull)


def test_iop_fitted():
    ziop_JCR = zmiopc.iopmod('ziop', data, X, Y, Z, pstart=pstartziop)
    fitttedziop = zmiopc.iopfit(ziop_JCR)
    assert len(fitttedziop.responseordered) == len(fitttedziop.responsefull)


# MiOP Examples
dataeu = pd.read_stata(os.getcwd()+"/data/EUKnowledge.dta")

Y2 = ["EU_support_ET"]
X2 = ['Xenophobia', 'discuss_politics']
Z2 = ['discuss_politics', 'EU_Know_obj']
yvar = np.unique(dataeu[Y2])


def test_miopc_coefs():
    miopc_EU = zmiopc.iopcmod('miopc',
                              dataeu, X2, Y2, Z2)
    assert len(miopc_EU.coefs) == len(X2 + Z2) + len(yvar) + 1


def test_miop_coefs():
    miop_EU = zmiopc.iopmod('miop', dataeu, X2, Y2, Z2)
    assert len(miop_EU.coefs) == len(X2 + Z2) + len(yvar)



import pandas as pd
from idcempy import gimnl
import os

DAT = pd.read_stata(
    os.getcwd()+"/data/replicationdata.dta", convert_categoricals=False)

x = ['educ', 'party7', 'agegroup2']
z = ['educ', 'agegroup2']
y = ['vote_turn']

order = [0, 1, 2]
second_order = [1, 0, 2]
torder = [2, 1, 0]

binflatecat = "baseline"
sinflatecat = "second"
tinflatecat = "third"


def test_gimnls():
    model = gimnl.gimnlmod(DAT, x, y, z, order, binflatecat)
    smodel = gimnl.gimnlmod(DAT, x, y, z, second_order, sinflatecat)
    tmodel = gimnl.gimnlmod(DAT, x, y, z, torder, tinflatecat)
    assert model.modeltype == smodel.modeltype == tmodel.modeltype



from os import path

from setuptools import find_packages, setup

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="idcempy",
    packages=find_packages(where="idcempy"),
    version="0.0.6",
    license='GPLv3',
    description="Inflated Discrete Choice Models",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Nguyen Huynh, Sergio Bejar, Vineeta Yadav, Bumba Mukherjee",
    author_email="nguyenhuynh831@gmail.com",
    url="https://github.com/hknd23/idcempy",
    keywords=["Inflated", "Mixture", "Ordered Probit", "Multinomial Logit"],
    install_requires=["scipy>=1.4.1", "numpy", "pandas"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3.7",
        "Intended Audience :: Science/Research",
    ],
)
